
/******************************************************************************
* MODULE     : generic.hh
* DESCRIPTION: Abstract types
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __GENERIC_HH
#define __GENERIC_HH
#include <gc.h>
#include <stdlib.h>
#include <iostream>
#include <config.hh>
using std::istream;
using std::ostream;
using std::cin;
using std::cout;
using std::cerr;
namespace mmx {

enum print_format {
  blank, stroke, indent, unindent, lf, pritty, normal, hrule
};
ostream& operator << (ostream& out, print_format x);

class generic;

#define UNKNOWN_TYPE         0
#define NIL_TYPE             1
#define STRING_TYPE          2
#define SYMBOL_TYPE          3
#define MIRROR_TYPE          4
#define ALIAS_TYPE           5
#define ARRAY_TYPE           6
#define LIST_TYPE            7
#define TABLE_TYPE           8
#define XTABLE_TYPE          9
#define FUNCTION_TYPE       10
#define ERROR_TYPE          11
#define SINGLE_TYPE         12
#define PRINTER_TYPE        13
#define VALUE_TYPE          14
#define LITERAL_INT_TYPE    15
#define LITERAL_FLOAT_TYPE  16
#define LITERAL_STRING_TYPE 17

#define EVALUATOR_TYPE      18
#define KEYWORD_TYPE        19
#define QUANTOR_TYPE        20
#define PROOF_TYPE          21
#define CONVERTER_TYPE      22

struct generic_rep {
  int begin, end;
  inline generic_rep (int b=-1, int e=-1): begin (b), end (e) {}
  inline generic_rep (generic g);
  inline virtual ~generic_rep () {}

  virtual int Type ();
  virtual int Hash ();
  virtual bool Equal (generic g);
  virtual generic Copy ();
  virtual int Arity ();
  virtual generic& Access (int i);
  virtual generic Range (int begin, int end);
  virtual generic& Car ();
  virtual generic Cdr ();
  virtual void Print (ostream& out);
};

struct generic {
  generic_rep* rep;

  inline generic (): rep (NULL) {}
  inline generic (generic_rep* rep2): rep (rep2) {}
  inline generic_rep* operator -> () { return rep; }
  inline bool operator == (generic g) {
    if (rep == NULL) return g.rep == NULL;
    return (rep == g.rep) || rep->Equal (g); }
  inline bool operator != (generic g) {
    if (rep == NULL) return g.rep != NULL;
    return (rep != g.rep) && (!rep->Equal (g)); }
  inline generic& operator [] (int i) { return rep->Access (i); }
  inline generic operator () (int beg, int end) {
    return rep->Range (beg, end); }
};

inline generic_rep::generic_rep (generic g) {
  if (g.rep != NULL) { begin= g->begin; end= g->end; } }
inline bool nil (generic g) { return g.rep==NULL; }
inline int gentype (generic g) {
  if (nil (g.rep)) return NIL_TYPE;
  return g.rep->Type (); }
inline int hash (generic g) {
  if (nil (g.rep)) return 0;
  return g.rep->Hash (); }
inline int N (generic g) {
  if (nil (g.rep)) return 0;
  return g.rep->Arity (); }
inline generic& car (generic g) { return g.rep->Car (); }
inline generic cdr (generic g) { return g.rep->Cdr (); }
inline ostream& operator << (ostream& out, generic g) {
  if (nil (g)) return out << "()";
  g.rep->Print (out);
  return out; }
void print_entries (ostream& out, generic g, bool flag= false);
generic& S (generic g);
generic extend (generic g, generic with); // really generic& g

/* should be declared as volatile, but not recognized by all g++ compilers */
void fatal_error (char* message, char* routine, char* file);
void type_mismatch (generic g, char* expected_type);
void type_mismatch (generic g, char* type, char* rout, char* file);

inline void* operator new   (register size_t sz) { return GC_MALLOC (sz); }
inline void* operator new[] (register size_t sz) { return GC_MALLOC (sz); }
inline void operator delete (register void* ptr)   { (void) ptr; }
inline void operator delete[] (register void* ptr) { (void) ptr; }

inline bool is_string (generic g) {
  return (!nil (g)) && (gentype (g) == STRING_TYPE); }
inline bool is_symbol (generic g) {
  return (!nil (g)) && (gentype (g) == SYMBOL_TYPE); }
inline bool is_list (generic g) {
  return nil(g) || (gentype(g) == LIST_TYPE); }
inline bool is_table (generic g) {
  return (!nil (g)) && (gentype (g) == TABLE_TYPE); }
inline bool is_gentable (generic g) {
  return (!nil (g)) &&
    (gentype (g) >= TABLE_TYPE) && (gentype(g) <= XTABLE_TYPE); }
inline bool is_function (generic g) {
  return (!nil (g)) && (gentype (g) == FUNCTION_TYPE); }
inline bool is_value (generic g) {
  return (!nil (g)) && (gentype (g) == VALUE_TYPE); }
inline bool is_single (generic g) {
  return (!nil (g)) && (gentype (g) == SINGLE_TYPE); }
inline bool is_literal (generic g) {
  return (!nil (g)) &&
    (gentype (g) >= LITERAL_INT_TYPE) &&
    (gentype (g) <= LITERAL_STRING_TYPE); }

inline bool is_func (generic g, generic s) {
  return is_list (g) && (N(g) > 0) && (car(g) == s); }
inline bool is_func (generic g, generic s, int n) {
  return is_list (g) && (N(g) == (n+1)) && (car(g) == s); }

}
#endif // __GENERIC_HH
